@@ -1,67 +1,73 @@
-# encoding: UTF-8
-# Copyright (C) 2003-2016, Stefan Schwarzer <sschwarzer@sschwarzer.net>
+# Copyright (C) 2003-2021, Stefan Schwarzer <sschwarzer@sschwarzer.net>
 # and ftputil contributors (see `doc/contributors.txt`)
 # See the file LICENSE for licensing terms.
 
-# Execute tests on a real FTP server (other tests use a mock server).
+# Execute tests on a real FTP server (other tests use mock code).
 #
-# This test writes some files and directories on the local client and
-# the remote server. You'll need write access in the login directory.
-# This test can take a few minutes because it has to wait to test the
-# timezone calculation.
-
-from __future__ import absolute_import
-from __future__ import unicode_literals
+# These tests write some files and directories on the local client and the
+# remote server. You'll need write access in the login directory. These tests
+# can take a few minutes because they have to wait to test the timezone
+# calculation.
 
+import datetime
 import ftplib
 import functools
 import gc
 import operator
 import os
+import pathlib
 import time
 import stat
 
 import pytest
 
-import ftputil.compat
 import ftputil.error
 import ftputil.file_transfer
+import ftputil.path_encoding
 import ftputil.session
 import ftputil.stat_cache
 
 import test
 
 
-def utc_local_time_shift():
+def expected_time_shift():
     """
-    Return the expected time shift in seconds assuming the server
-    uses UTC in its listings and the client uses local time.
+    Return the expected time shift in seconds.
 
-    This is needed because Pure-FTPd meanwhile seems to insist that
-    the displayed time for files is in UTC.
+    Unfortunately, the calculation may depend on the timezone of the server,
+    i. e. the timezone used in directory listings coming from the server.
+
+    So, depending on your test environment, you may need to change this
+    function for your environment. _If_ you need an adapted
+    `expected_time_shift`, please contact me (my e-mail address is in the
+    ftputil documentation).
+
+    In my particular case, I use Pure-FTPd as FTP server for the integration
+    tests. At some point, it returned listings in the local timezone of the
+    server, later it used UTC time, and now it uses the local timezone again.
+    I wasn't able to find out why or how I can control this.
     """
-    utc_tuple = time.gmtime()
-    localtime_tuple = time.localtime()
-    # To calculate the correct times shift, we need to ignore the
-    # DST component in the localtime tuple, i. e. set it to 0.
-    localtime_tuple = localtime_tuple[:-1] + (0,)
-    time_shift_in_seconds = (time.mktime(utc_tuple) -
-                             time.mktime(localtime_tuple))
-    # To be safe, round the above value to units of 3600 s (1 hour).
-    return round(time_shift_in_seconds / 3600.0) * 3600
+    raw_time_shift = (datetime.datetime.now() - datetime.datetime.utcnow()).seconds
+    # To be safe, round the above value to units of 900 s (1/4 hours).
+    return round(raw_time_shift / 900.0) * 900
 
-# Difference between local times of server and client. If 0.0, server
-# and client use the same timezone.
-#EXPECTED_TIME_SHIFT = utc_local_time_shift()
-# Pure-FTPd seems to have changed its mind (see docstring of
-# `utc_local_time_shift`).
+
+# The containerized PureFTPd seems to use UTC.
 EXPECTED_TIME_SHIFT = 0.0
 
 
-class Cleaner(object):
+# Set when starting container
+PORT = 2121
+
+DEFAULT_SESSION_FACTORY = ftputil.session.session_factory(
+    port=PORT, encrypt_data_channel=False
+)
+
+
+class Cleaner:
     """
-    This class helps remove directories and files which might
-    otherwise be left behind if a test fails in unexpected ways.
+    This class helps remove directories and files which might otherwise be left
+    behind if a test fails in unexpected ways.
     """
 
     def __init__(self, host):
@@ -71,18 +77,21 @@ class Cleaner(object):
         self._ftp_items = []
 
     def add_dir(self, path):
-        """Schedule a directory with path `path` for removal."""
+        """
+        Schedule a directory with path `path` for removal.
+        """
         self._ftp_items.append(("d", self._host.path.abspath(path)))
 
     def add_file(self, path):
-        """Schedule a file with path `path` for removal."""
+        """
+        Schedule a file with path `path` for removal.
+        """
         self._ftp_items.append(("f", self._host.path.abspath(path)))
 
     def clean(self):
         """
-        Remove the directories and files previously remembered.
-        The removal works in reverse order of the scheduling with
-        `add_dir` and `add_file`.
+        Remove the directories and files previously remembered. The removal
+        works in reverse order of the scheduling with `add_dir` and `add_file`.
 
         Errors due to a removal are ignored.
         """
@@ -90,8 +99,8 @@ class Cleaner(object):
         for type_, path in reversed(self._ftp_items):
             try:
                 if type_ == "d":
-                    # If something goes wrong in `rmtree` we might
-                    # leave a mess behind.
+                    # If something goes wrong in `rmtree` we might leave a mess
+                    # behind.
                     self._host.rmtree(path)
                 elif type_ == "f":
                     # Minor mess if `remove` fails
@@ -100,13 +109,13 @@ class Cleaner(object):
                 pass
 
 
-class RealFTPTest(object):
-
+class RealFTPTest:
     def setup_method(self, method):
         # Server, username, password.
-        self.login_data = ("localhost", "ftptest",
-                            "d605581757de5eb56d568a4419f4126e")
-        self.host = ftputil.FTPHost(*self.login_data)
+        self.login_data = ("localhost", "ftptest", "dummy")
+        self.host = ftputil.FTPHost(
+            *self.login_data, session_factory=DEFAULT_SESSION_FACTORY
+        )
         self.cleaner = Cleaner(self.host)
 
     def teardown_method(self, method):
@@ -120,18 +129,19 @@ class RealFTPTest(object):
         """Create a file on the FTP host."""
         self.cleaner.add_file(path)
         with self.host.open(path, "wb") as file_:
-            # Write something. Otherwise the FTP server might not update
-            # the time of last modification if the file existed before.
+            # Write something. Otherwise the FTP server might not update the
+            # time of last modification if the file existed before.
             file_.write(b"\n")
 
     def make_local_file(self):
-        """Create a file on the local host (= on the client side)."""
+        """
+        Create a file on the local host (= on the client side).
+        """
         with open("_local_file_", "wb") as fobj:
             fobj.write(b"abc\x12\x34def\t")
 
 
 class TestMkdir(RealFTPTest):
-
     def test_mkdir_rmdir(self):
         host = self.host
         dir_name = "_testdir_"
@@ -154,8 +164,7 @@ class TestMkdir(RealFTPTest):
             try:
                 host.remove(dir_name)
             except ftputil.error.PermanentError as exc:
-                assert str(exc).startswith(
-                         "remove/unlink can only delete files")
+                assert str(exc).startswith("remove/unlink can only delete files")
             else:
                 pytest.fail("we shouldn't have come here")
         finally:
@@ -209,7 +218,7 @@ class TestMkdir(RealFTPTest):
     def test_makedirs_of_existing_directory(self):
         host = self.host
         # The (chrooted) login directory
-        host.makedirs("/")
+        host.makedirs("/", exist_ok=True)
 
     def test_makedirs_with_file_in_the_way(self):
         host = self.host
@@ -240,15 +249,13 @@ class TestMkdir(RealFTPTest):
     def test_makedirs_with_writable_directory_at_end(self):
         host = self.host
         self.cleaner.add_dir("rootdir2/dir2")
-        # Preparation: `rootdir2` exists but is only writable by root.
-        # `dir2` is writable by regular ftp users. Both directories
-        # below should work.
-        host.makedirs("rootdir2/dir2")
+        # Preparation: `rootdir2` exists but is only writable by root. `dir2`
+        # is writable by regular ftp users. Both directories below should work.
+        host.makedirs("rootdir2/dir2", exist_ok=True)
         host.makedirs("rootdir2/dir2/dir3")
 
 
 class TestRemoval(RealFTPTest):
-
     def test_rmtree_without_error_handler(self):
         host = self.host
         # Build a tree.
@@ -279,8 +286,10 @@ class TestRemoval(RealFTPTest):
         self.make_remote_file("_dir1_/file1")
         # Prepare error "handler"
         log = []
+
         def error_handler(*args):
             log.append(args)
+
         # Try to remove a file as root "directory".
         host.rmtree("_dir1_/file1", ignore_errors=True, onerror=error_handler)
         assert log == []
@@ -336,7 +345,9 @@ class TestWalk(RealFTPTest):
     """
 
     def _walk_test(self, expected_result, **walk_kwargs):
-        """Walk the directory and test results."""
+        """
+        Walk the directory and test results.
+        """
         # Collect data using `walk`.
         actual_result = []
         for items in self.host.walk(**walk_kwargs):
@@ -349,143 +360,91 @@ class TestWalk(RealFTPTest):
     def test_walk_topdown(self):
         # Preparation: build tree in directory `walk_test`.
         expected_result = [
-          ("walk_test",
-           ["dir1", "dir2", "dir3"],
-           ["file4"]),
-          #
-          ("walk_test/dir1",
-           ["dir11", "dir12"],
-           []),
-          #
-          ("walk_test/dir1/dir11",
-           [],
-           []),
-          #
-          ("walk_test/dir1/dir12",
-           ["dir123"],
-           ["file121", "file122"]),
-          #
-          ("walk_test/dir1/dir12/dir123",
-           [],
-           ["file1234"]),
-          #
-          ("walk_test/dir2",
-           [],
-           []),
-          #
-          ("walk_test/dir3",
-           ["dir31", "dir32"],
-           ["file31", "file32"]),
-          #
-          ("walk_test/dir3/dir31",
-           [],
-           []),
-          ]
+            ("walk_test", ["dir1", "dir2", "dir3"], ["file4"]),
+            #
+            ("walk_test/dir1", ["dir11", "dir12"], []),
+            #
+            ("walk_test/dir1/dir11", [], []),
+            #
+            ("walk_test/dir1/dir12", ["dir123"], ["file121", "file122"]),
+            #
+            ("walk_test/dir1/dir12/dir123", [], ["file1234"]),
+            #
+            ("walk_test/dir2", [], []),
+            #
+            ("walk_test/dir3", ["dir31", "dir32"], ["file31", "file32"]),
+            #
+            ("walk_test/dir3/dir31", [], []),
+        ]
         self._walk_test(expected_result, top="walk_test")
 
     def test_walk_depth_first(self):
         # Preparation: build tree in directory `walk_test`
         expected_result = [
-          ("walk_test/dir1/dir11",
-           [],
-           []),
-          #
-          ("walk_test/dir1/dir12/dir123",
-           [],
-           ["file1234"]),
-          #
-          ("walk_test/dir1/dir12",
-           ["dir123"],
-           ["file121", "file122"]),
-          #
-          ("walk_test/dir1",
-           ["dir11", "dir12"],
-           []),
-          #
-          ("walk_test/dir2",
-           [],
-           []),
-          #
-          ("walk_test/dir3/dir31",
-           [],
-           []),
-          #
-          ("walk_test/dir3",
-           ["dir31", "dir32"],
-           ["file31", "file32"]),
-          #
-          ("walk_test",
-           ["dir1", "dir2", "dir3"],
-           ["file4"])
-          ]
+            ("walk_test/dir1/dir11", [], []),
+            #
+            ("walk_test/dir1/dir12/dir123", [], ["file1234"]),
+            #
+            ("walk_test/dir1/dir12", ["dir123"], ["file121", "file122"]),
+            #
+            ("walk_test/dir1", ["dir11", "dir12"], []),
+            #
+            ("walk_test/dir2", [], []),
+            #
+            ("walk_test/dir3/dir31", [], []),
+            #
+            ("walk_test/dir3", ["dir31", "dir32"], ["file31", "file32"]),
+            #
+            ("walk_test", ["dir1", "dir2", "dir3"], ["file4"]),
+        ]
         self._walk_test(expected_result, top="walk_test", topdown=False)
 
     def test_walk_following_links(self):
         # Preparation: build tree in directory `walk_test`.
         expected_result = [
-          ("walk_test",
-           ["dir1", "dir2", "dir3"],
-           ["file4"]),
-          #
-          ("walk_test/dir1",
-           ["dir11", "dir12"],
-           []),
-          #
-          ("walk_test/dir1/dir11",
-           [],
-           []),
-          #
-          ("walk_test/dir1/dir12",
-           ["dir123"],
-           ["file121", "file122"]),
-          #
-          ("walk_test/dir1/dir12/dir123",
-           [],
-           ["file1234"]),
-          #
-          ("walk_test/dir2",
-           [],
-           []),
-          #
-          ("walk_test/dir3",
-           ["dir31", "dir32"],
-           ["file31", "file32"]),
-          #
-          ("walk_test/dir3/dir31",
-           [],
-           []),
-          #
-          ("walk_test/dir3/dir32",
-           [],
-           ["file1234"]),
-          ]
+            ("walk_test", ["dir1", "dir2", "dir3"], ["file4"]),
+            #
+            ("walk_test/dir1", ["dir11", "dir12"], []),
+            #
+            ("walk_test/dir1/dir11", [], []),
+            #
+            ("walk_test/dir1/dir12", ["dir123"], ["file121", "file122"]),
+            #
+            ("walk_test/dir1/dir12/dir123", [], ["file1234"]),
+            #
+            ("walk_test/dir2", [], []),
+            #
+            ("walk_test/dir3", ["dir31", "dir32"], ["file31", "file32"]),
+            #
+            ("walk_test/dir3/dir31", [], []),
+            #
+            ("walk_test/dir3/dir32", [], ["file1234"]),
+        ]
         self._walk_test(expected_result, top="walk_test", followlinks=True)
 
 
 class TestRename(RealFTPTest):
-
     def test_rename(self):
         host = self.host
         # Make sure the target of the renaming operation is removed.
         self.cleaner.add_file("_testfile2_")
         self.make_remote_file("_testfile1_")
-        host.rename("_testfile1_", "_testfile2_")
+        host.rename(pathlib.Path("_testfile1_"), "_testfile2_")
         assert not host.path.exists("_testfile1_")
-        assert host.path.exists("_testfile2_")
+        assert host.path.exists(pathlib.Path("_testfile2_"))
 
     def test_rename_with_spaces_in_directory(self):
         host = self.host
         dir_name = "_dir with spaces_"
         self.cleaner.add_dir(dir_name)
-        host.mkdir(dir_name)
+        host.mkdir(pathlib.Path(dir_name))
         self.make_remote_file(dir_name + "/testfile1")
-        host.rename(dir_name + "/testfile1", dir_name + "/testfile2")
+        host.rename(dir_name + "/testfile1", pathlib.Path(dir_name + "/testfile2"))
         assert not host.path.exists(dir_name + "/testfile1")
         assert host.path.exists(dir_name + "/testfile2")
 
 
 class TestStat(RealFTPTest):
-
     def test_stat(self):
         host = self.host
         dir_name = "_testdir_"
@@ -498,29 +457,34 @@ class TestStat(RealFTPTest):
         # Do some stats
         # - dir
         dir_stat = host.stat(dir_name)
-        assert isinstance(dir_stat._st_name, ftputil.compat.unicode_type)
+        assert isinstance(dir_stat._st_name, str)
         assert host.listdir(dir_name) == ["_nonempty_"]
         assert host.path.isdir(dir_name)
         assert not host.path.isfile(dir_name)
         assert not host.path.islink(dir_name)
         # - file
         file_stat = host.stat(file_name)
-        assert isinstance(file_stat._st_name, ftputil.compat.unicode_type)
+        assert isinstance(file_stat._st_name, str)
         assert not host.path.isdir(file_name)
         assert host.path.isfile(file_name)
         assert not host.path.islink(file_name)
         assert host.path.getsize(file_name) == 9
-        # - file's modification time; allow up to two minutes difference
+        # - file's modification time
         host.synchronize_times()
+        #  The returned server mtime is supposed to be converted to UTC, so
+        #  there should be only a small difference between server and client
+        #  time. Arbitrarily allow two minutes here to account for limited time
+        #  precision from parsing the directory.
         server_mtime = host.path.getmtime(file_name)
-        client_mtime = time.mktime(time.localtime())
-        calculated_time_shift = server_mtime - client_mtime
-        assert not abs(calculated_time_shift-host.time_shift()) > 120
+        client_mtime = (
+            datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).timestamp()
+        )
+        assert not (client_mtime - server_mtime > 120)
 
     def test_issomething_for_nonexistent_directory(self):
         host = self.host
-        # Check if we get the right results if even the containing
-        # directory doesn't exist (see ticket #66).
+        # Check if we get the right results if even the containing directory
+        # doesn't exist (see ticket #66).
         nonexistent_path = "/nonexistent/nonexistent"
         assert not host.path.isdir(nonexistent_path)
         assert not host.path.isfile(nonexistent_path)
@@ -530,16 +494,19 @@ class TestStat(RealFTPTest):
         # Test for ticket #39.
         host = self.host
         broken_link_name = os.path.join("dir_with_broken_link", "nonexistent")
-        assert (host.lstat(broken_link_name)._st_target ==
-                "../nonexistent/nonexistent")
+        assert host.lstat(broken_link_name)._st_target == "../nonexistent/nonexistent"
         assert not host.path.isdir(broken_link_name)
         assert not host.path.isfile(broken_link_name)
         assert host.path.islink(broken_link_name)
 
     def test_concurrent_access(self):
         self.make_remote_file("_testfile_")
-        with ftputil.FTPHost(*self.login_data) as host1:
-            with ftputil.FTPHost(*self.login_data) as host2:
+        with ftputil.FTPHost(
+            *self.login_data, session_factory=DEFAULT_SESSION_FACTORY
+        ) as host1:
+            with ftputil.FTPHost(
+                *self.login_data, session_factory=DEFAULT_SESSION_FACTORY
+            ) as host2:
                 stat_result1 = host1.stat("_testfile_")
                 stat_result2 = host2.stat("_testfile_")
                 assert stat_result1 == stat_result2
@@ -557,7 +524,9 @@ class TestStat(RealFTPTest):
                     host1.stat("_testfile_")
 
     def test_cache_auto_resizing(self):
-        """Test if the cache is resized appropriately."""
+        """
+        Test if the cache is resized appropriately.
+        """
         host = self.host
         cache = host.stat_cache._cache
         # Make sure the cache size isn't adjusted towards smaller values.
@@ -566,16 +535,17 @@ class TestStat(RealFTPTest):
         # Make the cache very small initially and see if it gets resized.
         cache.size = 2
         entries = host.listdir("walk_test")
-        # The adjusted cache size should be larger or equal to the
-        # number of items in `walk_test` and its parent directory. The
-        # latter is read implicitly upon `listdir`'s `isdir` call.
-        expected_min_cache_size = max(len(host.listdir(host.curdir)),
-                                      len(entries))
+        # The adjusted cache size should be larger or equal to the number of
+        # items in `walk_test` and its parent directory. The latter is read
+        # implicitly upon `listdir`'s `isdir` call.
+        expected_min_cache_size = max(len(host.listdir(host.curdir)), len(entries))
         assert cache.size >= expected_min_cache_size
 
 
 class TestUploadAndDownload(RealFTPTest):
-    """Test upload and download (including time shift test)."""
+    """
+    Test upload and download (including time shift test).
+    """
 
     def test_time_shift(self):
         self.host.synchronize_times()
@@ -589,8 +559,8 @@ class TestUploadAndDownload(RealFTPTest):
         remote_file = "_remote_file_"
         # Make local file to upload.
         self.make_local_file()
-        # Wait, else small time differences between client and server
-        # actually could trigger the update.
+        # Wait, else small time differences between client and server actually
+        # could trigger the update.
         time.sleep(65)
         try:
             self.cleaner.add_file(remote_file)
@@ -619,10 +589,10 @@ class TestUploadAndDownload(RealFTPTest):
         downloaded = host.download_if_newer(remote_file, local_file)
         assert downloaded is True
         try:
-            # If the remote file, taking the datetime precision into
-            # account, _might_ be newer, the file will be downloaded
-            # again. To prevent this, wait a bit over a minute (the
-            # remote precision), then "touch" the local file.
+            # If the remote file, taking the datetime precision into account,
+            # _might_ be newer, the file will be downloaded again. To prevent
+            # this, wait a bit over a minute (the remote precision), then
+            # "touch" the local file.
             time.sleep(65)
             # Create empty file.
             with open(local_file, "w") as fobj:
@@ -632,9 +602,9 @@ class TestUploadAndDownload(RealFTPTest):
             assert downloaded is False
             # Re-make the remote file.
             self.make_remote_file(remote_file)
-            # Local file is present but possibly older (taking the
-            # possible deviation because of the precision into account),
-            # so should download.
+            # Local file is present but possibly older (taking the possible
+            # deviation because of the precision into account), so should
+            # download.
             downloaded = host.download_if_newer(remote_file, local_file)
             assert downloaded is True
         finally:
@@ -643,7 +613,7 @@ class TestUploadAndDownload(RealFTPTest):
 
     def test_callback_with_transfer(self):
         host = self.host
-        FILE_NAME = "debian-keyring.tar.gz"
+        FILE_NAME = "large_file"
         # Default chunk size as in `FTPHost.copyfileobj`
         MAX_COPY_CHUNK_SIZE = ftputil.file_transfer.MAX_COPY_CHUNK_SIZE
         file_size = host.path.getsize(FILE_NAME)
@@ -652,8 +622,10 @@ class TestUploadAndDownload(RealFTPTest):
         chunk_count += 1
         # Define a callback that just collects all data passed to it.
         transferred_chunks_list = []
+
         def test_callback(chunk):
             transferred_chunks_list.append(chunk)
+
         try:
             host.download(FILE_NAME, FILE_NAME, callback=test_callback)
             # Construct a list of data chunks we expect.
@@ -672,7 +644,6 @@ class TestUploadAndDownload(RealFTPTest):
 
 
 class TestFTPFiles(RealFTPTest):
-
     def test_only_closed_children(self):
         REMOTE_FILE_NAME = "CONTENTS"
         host = self.host
@@ -695,6 +666,7 @@ class TestFTPFiles(RealFTPTest):
         # Monkey-patch file to simulate an FTP server timeout below.
         def timed_out_pwd():
             raise ftplib.error_temp("simulated timeout")
+
         file_obj1._host._session.pwd = timed_out_pwd
         # Try to get a file - which shouldn't be the timed-out file.
         with host.open(REMOTE_FILE_NAME, "rb") as file_obj2:
@@ -713,6 +685,7 @@ class TestFTPFiles(RealFTPTest):
         # Monkey-patch file to simulate an FTP server timeout below.
         def timed_out_pwd():
             raise ftplib.error_reply("delayed 226 reply")
+
         file_obj1._host._session.pwd = timed_out_pwd
         # Try to get a file - which shouldn't be the timed-out file.
         with host.open(REMOTE_FILE_NAME, "rb") as file_obj2:
@@ -724,27 +697,40 @@ class TestFTPFiles(RealFTPTest):
 
 
 class TestChmod(RealFTPTest):
-
     def assert_mode(self, path, expected_mode):
         """
-        Return an integer containing the allowed bits in the mode
-        change command.
+        Return an integer containing the allowed bits in the mode change
+        command.
 
         The `FTPHost` object to test against is `self.host`.
         """
         full_mode = self.host.stat(path).st_mode
-        # Remove flags we can't set via `chmod`.
-        # Allowed flags according to Python documentation
-        # https://docs.python.org/library/stat.html
-        allowed_flags = [stat.S_ISUID, stat.S_ISGID, stat.S_ENFMT,
-          stat.S_ISVTX, stat.S_IREAD, stat.S_IWRITE, stat.S_IEXEC,
-          stat.S_IRWXU, stat.S_IRUSR, stat.S_IWUSR, stat.S_IXUSR,
-          stat.S_IRWXG, stat.S_IRGRP, stat.S_IWGRP, stat.S_IXGRP,
-          stat.S_IRWXO, stat.S_IROTH, stat.S_IWOTH, stat.S_IXOTH]
+        # Remove flags we can't set via `chmod`. Allowed flags according to
+        # Python documentation: https://docs.python.org/library/stat.html
+        allowed_flags = [
+            stat.S_ISUID,
+            stat.S_ISGID,
+            stat.S_ENFMT,
+            stat.S_ISVTX,
+            stat.S_IREAD,
+            stat.S_IWRITE,
+            stat.S_IEXEC,
+            stat.S_IRWXU,
+            stat.S_IRUSR,
+            stat.S_IWUSR,
+            stat.S_IXUSR,
+            stat.S_IRWXG,
+            stat.S_IRGRP,
+            stat.S_IWGRP,
+            stat.S_IXGRP,
+            stat.S_IRWXO,
+            stat.S_IROTH,
+            stat.S_IWOTH,
+            stat.S_IXOTH,
+        ]
         allowed_mask = functools.reduce(operator.or_, allowed_flags)
         mode = full_mode & allowed_mask
-        assert mode == expected_mode, (
-                 "mode {0:o} != {1:o}".format(mode, expected_mode))
+        assert mode == expected_mode, "mode {0:o} != {1:o}".format(mode, expected_mode)
 
     def test_chmod_existing_directory(self):
         host = self.host
@@ -797,7 +783,7 @@ class TestRestArgument(RealFTPTest):
     TEST_FILE_NAME = "rest_test"
 
     def setup_method(self, method):
-        super(TestRestArgument, self).setup_method(method)
+        super().setup_method(method)
         # Write test file.
         with self.host.open(self.TEST_FILE_NAME, "wb") as fobj:
             fobj.write(b"abcdefghijkl")
@@ -805,8 +791,8 @@ class TestRestArgument(RealFTPTest):
 
     def test_for_reading(self):
         """
-        If a `rest` argument is passed to `open`, the following read
-        operation should start at the byte given by `rest`.
+        If a `rest` argument is passed to `open`, the following read operation
+        should start at the byte given by `rest`.
         """
         with self.host.open(self.TEST_FILE_NAME, "rb", rest=3) as fobj:
             data = fobj.read()
@@ -814,8 +800,8 @@ class TestRestArgument(RealFTPTest):
 
     def test_for_writing(self):
         """
-        If a `rest` argument is passed to `open`, the following write
-        operation should start writing at the byte given by `rest`.
+        If a `rest` argument is passed to `open`, the following write operation
+        should start writing at the byte given by `rest`.
         """
         with self.host.open(self.TEST_FILE_NAME, "wb", rest=3) as fobj:
             fobj.write(b"123")
@@ -825,50 +811,47 @@ class TestRestArgument(RealFTPTest):
 
     def test_invalid_read_from_text_file(self):
         """
-        If the `rest` argument is used for reading from a text file,
-        a `CommandNotImplementedError` should be raised.
+        If the `rest` argument is used for reading from a text file, a
+        `CommandNotImplementedError` should be raised.
         """
         with pytest.raises(ftputil.error.CommandNotImplementedError):
             self.host.open(self.TEST_FILE_NAME, "r", rest=3)
 
     def test_invalid_write_to_text_file(self):
         """
-        If the `rest` argument is used for reading from a text file,
-        a `CommandNotImplementedError` should be raised.
+        If the `rest` argument is used for reading from a text file, a
+        `CommandNotImplementedError` should be raised.
         """
         with pytest.raises(ftputil.error.CommandNotImplementedError):
             self.host.open(self.TEST_FILE_NAME, "w", rest=3)
 
-    # There are no tests for reading and writing beyond the end of a
-    # file. For example, if the remote file is 10 bytes long and
-    # `open(remote_file, "rb", rest=100)` is used, the server may
-    # return an error status code or not.
+    # There are no tests for reading and writing beyond the end of a file. For
+    # example, if the remote file is 10 bytes long and
+    # `open(remote_file, "rb", rest=100)` is used, the server may return an
+    # error status code or not.
     #
-    # The server I use for testing returns a 554 status when
-    # attempting to _read_ beyond the end of the file. On the other
-    # hand, if attempting to _write_ beyond the end of the file, the
-    # server accepts the request, but starts writing after the end of
-    # the file, i. e. appends to the file.
+    # The server I use for testing returns a 554 status when attempting to
+    # _read_ beyond the end of the file. On the other hand, if attempting to
+    # _write_ beyond the end of the file, the server accepts the request, but
+    # starts writing after the end of the file, i. e. appends to the file.
     #
-    # Instead of expecting certain responses that may differ between
-    # server implementations, I leave the bahavior for too large
-    # `rest` arguments undefined. In practice, this shouldn't be a
-    # problem because the `rest` argument should only be used for
-    # error recovery, and in this case a valid byte count for the
-    # `rest` argument should be known.
+    # Instead of expecting certain responses that may differ between server
+    # implementations, I leave the bahavior for too large `rest` arguments
+    # undefined. In practice, this shouldn't be a problem because the `rest`
+    # argument should only be used for error recovery, and in this case a valid
+    # byte count for the `rest` argument should be known.
 
 
 class TestOther(RealFTPTest):
-
     def test_open_for_reading(self):
         # Test for issues #17 and #51,
         # http://ftputil.sschwarzer.net/trac/ticket/17 and
         # http://ftputil.sschwarzer.net/trac/ticket/51 .
-        file1 = self.host.open("debian-keyring.tar.gz", "rb")
+        file1 = self.host.open("large_file", "rb")
         time.sleep(1)
         # Depending on the FTP server, this might return a status code
-        # unexpected by `ftplib` or block the socket connection until
-        # a server-side timeout.
+        # unexpected by `ftplib` or block the socket connection until a
+        # server-side timeout.
         file1.close()
 
     def test_subsequent_reading(self):
@@ -881,29 +864,41 @@ class TestOther(RealFTPTest):
         assert file1._session is file2._session
 
     def test_names_with_spaces(self):
-        # Test if directories and files with spaces in their names
-        # can be used.
+        # Test if directories and files with spaces in their names can be used.
         host = self.host
         assert host.path.isdir("dir with spaces")
-        assert (host.listdir("dir with spaces") ==
-                ["second dir", "some file", "some_file"])
+        assert host.listdir("dir with spaces") == [
+            "second dir",
+            "some file",
+            "some_file",
+        ]
         assert host.path.isdir("dir with spaces/second dir")
         assert host.path.isfile("dir with spaces/some_file")
         assert host.path.isfile("dir with spaces/some file")
 
     def test_synchronize_times_without_write_access(self):
-        """Test failing synchronization because of non-writable directory."""
+        """
+        Test failing synchronization because of non-writable directory.
+        """
         host = self.host
         # This isn't writable by the ftp account the tests are run under.
         host.chdir("rootdir1")
         with pytest.raises(ftputil.error.TimeShiftError):
             host.synchronize_times()
 
+    def test_encoding(self):
+        """
+        Test setting the path encoding.
+        """
+        for encoding in ["latin-1", "UTF-8"]:
+            factory = ftputil.session.session_factory(port=PORT, encoding=encoding)
+            session = factory(*self.login_data)
+            assert session.encoding == encoding
+
     def test_listdir_with_non_ascii_byte_string(self):
         """
-        `listdir` should accept byte strings with non-ASCII
-        characters and return non-ASCII characters in directory or
-        file names.
+        `listdir` should accept byte strings with non-ASCII characters and
+        return non-ASCII characters in directory or file names.
         """
         host = self.host
         path = "äbc".encode("UTF-8")
@@ -913,46 +908,55 @@ class TestOther(RealFTPTest):
 
     def test_listdir_with_non_ascii_unicode_string(self):
         """
-        `listdir` should accept unicode strings with non-ASCII
-        characters and return non-ASCII characters in directory or
-        file names.
+        `listdir` should accept unicode strings with non-ASCII characters and
+        return non-ASCII characters in directory or file names.
         """
         host = self.host
-        # `ftplib` under Python 3 only works correctly if the unicode
-        # strings are decoded from latin1. Under Python 2, ftputil
-        # is supposed to provide a compatible interface.
-        path = "äbc".encode("UTF-8").decode("latin1")
+        # `ftplib` under Python 3 only works correctly if the unicode strings
+        # are decoded from the `ftplib` default encoding. For Python 3.9 and up
+        # the `encode`/`decode` combination is a no-op.
+        path = "äbc".encode("UTF-8").decode(
+            ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING
+        )
         names = host.listdir(path)
         assert names[0] == "file1"
-        assert names[1] == "file1_ö".encode("UTF-8").decode("latin1")
+        assert names[1] == "file1_ö".encode("UTF-8").decode(
+            ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING
+        )
 
-    def test_path_with_non_latin1_unicode_string(self):
+    @pytest.mark.skipif(
+        ftputil.path_encoding.RUNNING_UNDER_PY39_AND_UP,
+        reason="test applies only to `FTPHost` objects using 'latin-1' path encoding",
+    )
+    def test_path_with_non_latin1_unicode_string_below_python_3_9(self):
         """
         ftputil operations shouldn't accept file paths with non-latin1
         characters.
         """
-        # Use some musical symbols. These are certainly not latin1.
+        # Use some musical symbols. These are certainly not latin1. ;-)
         path = "𝄞𝄢"
-        # `UnicodeEncodeError` is also the exception that `ftplib`
-        # raises if it gets a non-latin1 path.
+        # `UnicodeEncodeError` is also the exception that `ftplib` raises if it
+        # gets a non-latin1 path.
         with pytest.raises(UnicodeEncodeError):
             self.host.mkdir(path)
 
     def test_list_a_option(self):
-        # For this test to pass, the server must _not_ list "hidden"
-        # files by default but instead only when the `LIST` `-a`
-        # option is used.
+        # For this test to pass, the server must _not_ list "hidden" files by
+        # default but instead only when the `LIST` `-a` option is used.
         host = self.host
-        assert host.use_list_a_option
-        directory_entries = host.listdir(host.curdir)
-        assert ".hidden" in directory_entries
-        host.use_list_a_option = False
+        assert not host.use_list_a_option
         directory_entries = host.listdir(host.curdir)
         assert ".hidden" not in directory_entries
+        # Switch on showing of hidden paths.
+        host.use_list_a_option = True
+        directory_entries = host.listdir(host.curdir)
+        assert ".hidden" in directory_entries
 
     def _make_objects_to_be_garbage_collected(self):
         for _ in range(10):
-            with ftputil.FTPHost(*self.login_data) as host:
+            with ftputil.FTPHost(
+                *self.login_data, session_factory=DEFAULT_SESSION_FACTORY
+            ) as host:
                 for _ in range(10):
                     unused_stat_result = host.stat("CONTENTS")
                     with host.open("CONTENTS") as fobj:
@@ -966,23 +970,3 @@ class TestOther(RealFTPTest):
         gc.collect()
         objects_after_test = len(gc.garbage)
         assert not objects_after_test - objects_before_test
-
-    @pytest.mark.skipif(
-      ftputil.compat.python_version > 2,
-      reason="test requires M2Crypto which only works on Python 2")
-    def test_m2crypto_session(self):
-        """
-        Test if a session with `M2Crypto.ftpslib.FTP_TLS` is set up
-        correctly and works with unicode input.
-        """
-        # See ticket #78.
-        #
-        # M2Crypto is only available for Python 2.
-        import M2Crypto
-        factory = ftputil.session.session_factory(
-                    base_class=M2Crypto.ftpslib.FTP_TLS,
-                    encrypt_data_channel=True)
-        with ftputil.FTPHost(*self.login_data, session_factory=factory) as host:
-            # Test if unicode argument works.
-            files = host.listdir(".")
-        assert "CONTENTS" in files
